// This file is a bit misnamed, as it has options for giveAwrl, too.
const flags = [
    'vrmlview', 'joyce',
    // debugging
    'color', 'commands', 'showall', 'showaux', 'algebra', 'vrml97'
];
function params(kids) {
    const ret = {};
    for (let i = 0, len = kids.length; i < len; i++) {
        const kid = kids[i];
        if (!(kid.tagName === 'PARAM'))
            continue;
        const name = kid.getAttribute('name');
        const value = kid.getAttribute('value');
        if (!(name && value))
            continue;
        ret[name] = value;
    }
    return ret;
}
const prim3d = /polyhedron|sphere|plane|face/;
function contains3d(params) {
    for (const value of Object.values(params)) {
        if (prim3d.test(value))
            return true;
    }
    return false;
}

// (At least some of) Joyce's pages have inline scripts that remove their
// own applet nodes, so we have to watch for them as they are added. Hence,
// this script runs at document_start, and watches for applets being added,
// setting up the joyceApplets structure as it goes
let joyceApplets = [];
const obs = new MutationObserver((mutationList) => {
    const results = [];
    for (let i = 0, len = mutationList.length; i < len; i++) {
        const change = mutationList[i];
        const results1 = [];
        for (let ref = change.addedNodes, i1 = 0, len1 = ref.length; i1 < len1; i1++) {
            const newGenericNode = ref[i1];
            const newNode = newGenericNode;
            const newParent = change.target;
            if (!(newNode.tagName === 'APPLET'))
                continue;
            const code = newNode.getAttribute('code');
            if (!(code === 'Geometry' || code === 'Geometry.class'))
                continue;
            let id = newParent.getAttribute('id');
            if (!id) {
                id = 'joyceApplet' + joyceApplets.length;
                newParent.setAttribute('id', id);
            }
            results1.push(joyceApplets.push({
                html: newNode.outerHTML,
                params: params(newNode.children),
                id,
                width: parseInt(newNode.getAttribute('width') ?? '200'),
                height: parseInt(newNode.getAttribute('height') ?? '200')
            }));
        }
        results.push(results1);
    }
    return results;
});
const observerConfig = { childList: true, subtree: true };
obs.observe(document.documentElement, observerConfig);
function addScriptTag(url, params = '', module = false) {
    return new Promise((resolve, reject) => {
        const script = document.createElement('script');
        script.addEventListener('load', resolve);
        script.addEventListener('error', reject);
        script.src = url;
        if (params)
            script.dataset.params = params;
        if (module)
            script.type = 'module';
        return document.head.appendChild(script);
    });
}
document.addEventListener("DOMContentLoaded", async () => {
    const config = await browser.storage.local.get(flags);
    if (!config.joyce)
        return;
    const finalJoyceApplets = document.querySelectorAll("applet[code='Geometry']");
    if (finalJoyceApplets.length) {
        joyceApplets = [];
        for (const jApp of finalJoyceApplets) {
            const jParent = jApp.parentNode;
            if (!jParent)
                continue;
            let id = jParent.getAttribute('id');
            if (!id) {
                id = 'joyceApplet' + joyceApplets.length;
                jParent.setAttribute('id', id);
            }
            joyceApplets.push({
                html: jApp.outerHTML,
                params: params(jApp.children),
                id,
                width: parseInt(jApp.getAttribute('width') ?? '200'),
                height: parseInt(jApp.getAttribute('height') ?? '200')
            });
        }
    }
    if (joyceApplets.length) {
        let use3d = false;
        for (let i2 = 0, len2 = joyceApplets.length; i2 < len2; i2++) {
            const jApp = joyceApplets[i2];
            if (contains3d(jApp.params)) {
                use3d = true;
                break;
            }
        }
        let codebase = browser.runtime.getURL('deps/GeoGebra/HTML5/5.0/');
        codebase += (use3d || config.algebra) ? 'web3d' : 'webSimple';
        const adapParams = { codebase, config, joyceApplets };
        const apars = JSON.stringify(adapParams);
        return addScriptTag(browser.runtime.getURL('deps/GeoGebra/deployggb.js')).then(() => {
            return addScriptTag(browser.runtime.getURL('adapptlet.js'), apars, true);
        });
    }
    return;
});
