var concatAssign = (lhs, rhs) => ((rhs?.[Symbol.isConcatSpreadable] ?? Array.isArray(rhs)) ? lhs.push.apply(lhs, rhs) : lhs.push(rhs), lhs);
var xor = (a, b) => (a ? !b && a : b);
// Useful constants
const rt2 = Math.sqrt(2);
const rth = rt2 / 2;
const phi = (1 + Math.sqrt(5)) / 2;
const ihp = 1 / phi;
const tau = 2 * Math.PI;
// Sadly needs to be early because we initialize the Dodecahedron as the dual of
// the icosahedron:
// Only add one direction of each, will auto reverse as well
const specialDuals = {
    Tetrahedron: 'Tetrahedron',
    Cube: 'Octahdron',
    Dodecahedron: 'Icosahedron',
    Cuboctahedron: 'Rhombic dodecahedron',
    'truncated Tetrahedron': 'triakis Tetrahedron',
    'truncated Cube': 'triakis Octahedron',
    'truncated Octahedron': 'tetrakis Cube',
    Rhombicuboctahedron: 'Deltoidal icositetrahedron',
    'truncated Cuboctahedron': 'Disdyakis dodecahedron',
    'snub Cube': 'pentagonal icositetrahedron',
    Icosidodecahedron: 'Rhombic triacontahedron',
    'truncated Dodecahedron': 'triakis Icosahedron',
    'truncated Icosahedron': 'pentakis Dodecahedron',
    Rhombicosidodecahedron: 'Deltoidal hexecontahedron',
    'truncated Icosidodecahedron': 'Disdyakis triacontahedron',
    'snub Dodecahedron': 'pentagonal hexecontahedron',
};
// All seeds are canonical, i.e., edges tangent to unit sphere
const icosahedron = {
    face: [[0, 1, 9], [0, 8, 1], [0, 4, 8], [0, 5, 4], [0, 9, 5],
        [4, 5, 2], [4, 2, 10], [4, 10, 8], [8, 10, 6], [8, 6, 1],
        [1, 6, 7], [1, 7, 9], [9, 7, 11], [9, 11, 5], [5, 11, 2],
        [3, 2, 11], [3, 10, 2], [3, 6, 10], [3, 7, 6], [3, 11, 7]],
    xyz: [[0, ihp, 1], [0, -ihp, 1], [0, ihp, -1], [0, -ihp, -1],
        [ihp, 1, 0], [-ihp, 1, 0], [ihp, -1, 0], [-ihp, -1, 0],
        [1, 0, ihp], [-1, 0, ihp], [1, 0, -ihp], [-1, 0, -ihp]],
    name: 'Icosahedron',
};
const polyCache = {
    '': { face: [], xyz: [] },
    T: {
        face: [[0, 2, 1], [0, 3, 2], [0, 1, 3], [1, 2, 3]],
        xyz: [[1, 1, 1], [1, -1, -1], [-1, 1, -1], [-1, -1, 1]],
        name: 'Tetrahedron',
    },
    O: {
        face: [[0, 2, 1], [0, 3, 2], [0, 4, 3], [0, 1, 4],
            [1, 5, 4], [1, 2, 5], [2, 3, 5], [3, 4, 5]],
        xyz: [[0, 0, rt2], [rt2, 0, 0], [0, rt2, 0], [-rt2, 0, 0], [0, -rt2, 0], [0, 0, -rt2]],
        name: 'Octahedron',
    },
    C: {
        face: [[0, 3, 2, 1], [0, 5, 4, 3], [0, 1, 6, 5], [1, 2, 7, 6], [2, 3, 4, 7], [4, 5, 6, 7]],
        xyz: [[rth, rth, rth], [-rth, rth, rth], [-rth, -rth, rth], [rth, -rth, rth],
            [rth, -rth, -rth], [rth, rth, -rth], [-rth, rth, -rth], [-rth, -rth, -rth]],
        name: 'Cube',
    },
    I: icosahedron,
    D: geomDual(icosahedron),
};
export function generateVRML(notation) {
    return outputVRML(notation, generatePoly(notation));
}
function generatePoly(notation) {
    return getStandardPoly(inform(standardize(notation)));
}
function getStandardPoly(notation) {
    if (notation in polyCache)
        return polyCache[notation];
    const [, op, arg, rest] = notation.match(/^(.)(\d*)(.*)$/) || [];
    const parent = getStandardPoly(rest);
    // may have created what we want by side effect
    if (notation in polyCache)
        return polyCache[notation];
    return dispatch(op, arg, parent, notation);
} // will do the caching
// Convenience tuple maker
function ð(...arg) { return arg; }
// Note we now allow numeric arguments on all of the basic operations,
// kis/truncate, join/ambo, and gyro/snub. Likely some of the operations
// we are taking as composite could have reasonable numeric versions, but
// there didn't seem to be any sensible way to propagate such an argument
// to the operations in their rewrites. In other words, the numeric-limited
// operations may not be composite, or at least not in the same way. So
// we have just left them as applying throughout the polyhedron.
const rawStandardizations = {
    P4$: 'C', A3$: 'O', Y3$: 'T', // Seed synonyms
    e: 'aa', b: 'ta', o: 'jj', m: 'kj', // abbreviations
    [String.raw `t(\d*)`]: 'dk$1d',
    [String.raw `a(\d*)`]: 'dj$1d', // dual operations
    [String.raw `s(\d*)`]: 'dg$1d',
    dd: '', rr: '', jd: 'j', gd: 'rgr', // absorption rules
    rd: 'dr', // these commute; others? If so, move 'r' in to cancel w/ seed
    // Remainder are all simplifications/unique selections for seeds:
    jY: 'dA', dT: 'T', gT: 'D', jT: 'C', dC: 'O', dO: 'C',
    dI: 'D', dD: 'I', rO: 'O', rC: 'C', rI: 'I', rD: 'D',
    jO: 'jC', jI: 'jD', gO: 'gC', gI: 'gD',
};
const standardizations = (() => { const results = []; for (const pat in rawStandardizations) {
    const rep = rawStandardizations[pat];
    results.push(ð(RegExp(pat, 'g'), rep));
} return results; })();
function standardize(notation) {
    let lastNotation = '';
    while (lastNotation != notation) { // iterate in case of rdrd, e.g.
        lastNotation = notation;
        for (const [pat, rep] of standardizations) {
            notation = notation.replace(pat, rep);
        }
    }
    return notation;
}
function orb(r, n, height, t = 0) {
    // A regular n-gon inscribed in a horizontal circle of radius r
    // at given height, rotated t*tau/n from standard position.
    const theta = tau / n;
    const rho = t * theta;
    const results1 = [];
    for (let i1 = 0, asc = 0 <= n; asc ? i1 < n : i1 > n; asc ? ++i1 : --i1) {
        const i = i1;
        results1.push([r * Math.cos(rho + i * theta), r * Math.sin(rho + i * theta), height]);
    }
    ;
    return results1;
}
function add(a, b) {
    return accumulate(copy(a), b);
}
const ngonalNames = { 3: 'triangular', 4: 'square', 5: 'pentagonal', 6: 'hexagonal',
    7: 'heptagonal', 8: 'octagonal', 9: 'enneagonal', 10: 'decagonal',
    12: 'dodecagonal' };
function ngonal(n) {
    return ngonalNames[n] ?? n + `-gonal`;
}
const seeds = {
    P: (n) => {
        if (!n)
            n = 3;
        const theta = tau / n;
        const halfEdge = Math.sin(theta / 2);
        const xyz = orb(1, n, halfEdge).concat(orb(1, n, -halfEdge));
        const face = [((s, e) => { let step = e > s ? 1 : -1; return Array.from({ length: Math.abs(e - s) + 1 }, (_, i) => s + i * step); })(n - 1, 0), ((s, e) => { let step = e > s ? 1 : -1; return Array.from({ length: Math.abs(e - s) }, (_, i) => s + i * step); })(n, 2 * n)]; // top and bottom
        for (let i2 = 0, asc1 = 0 <= n; asc1 ? i2 < n : i2 > n; asc1 ? ++i2 : --i2) {
            const i = i2;
            const ip1 = (i + 1) % n; // next vertex around
            face.push([i, ip1, ip1 + n, i + n]);
        }
        return ({ face, xyz, name: ngonal(n) + ` prism` });
    },
    A: (n) => {
        if (!n)
            n = 4;
        const theta = tau / n;
        let halfHeight = Math.sqrt(1 - 4 / (4 + 2 * Math.cos(theta / 2) - 2 * Math.cos(theta)));
        let faceRadius = Math.sqrt(1 - halfHeight * halfHeight);
        // Scale to put edge midpoints on unit sphere
        const f = mag([halfHeight, faceRadius * Math.cos(theta / 2), 0]);
        halfHeight /= f;
        faceRadius /= f;
        const xyz = orb(faceRadius, n, halfHeight)
            .concat(orb(faceRadius, n, -halfHeight, 0.5));
        const face = [((s, e) => { let step = e > s ? 1 : -1; return Array.from({ length: Math.abs(e - s) + 1 }, (_, i) => s + i * step); })(n - 1, 0), ((s, e) => { let step = e > s ? 1 : -1; return Array.from({ length: Math.abs(e - s) }, (_, i) => s + i * step); })(n, 2 * n)]; // top and bottom
        for (let i3 = 0, asc2 = 0 <= n; asc2 ? i3 < n : i3 > n; asc2 ? ++i3 : --i3) {
            const i = i3;
            face.push([i, (i + 1) % n, i + n]);
            face.push([i, i + n, n + (n + i - 1) % n]);
        }
        return ({ face, xyz, name: ngonal(n) + ` antiprism` });
    },
    Y: (n) => {
        if (!n)
            n = 4;
        // Canonical solution by Intelligenti Pauca and Ed Pegg, see
        // https://math.stackexchange.com/questions/2286628/canonical-pyramid-polynomials
        const theta = tau / n;
        const c = Math.cos(theta / 2);
        const baseRadius = Math.sqrt(2 / (c * (1 + c)));
        const depth = Math.sqrt((1 - c) / (1 + c));
        const height = 2 * Math.sqrt(1 / (1 - c * c));
        const xyz = orb(baseRadius, n, depth);
        const edgeMid2 = add(xyz[0], xyz[1]);
        xyz.push([0, 0, depth - height]);
        const face = (() => { const results2 = []; for (let i4 = 0, asc3 = 0 <= n; asc3 ? i4 < n : i4 > n; asc3 ? ++i4 : --i4) {
            const i = i4;
            results2.push([i, (i + 1) % n, n]);
        } return results2; })();
        face.unshift(((s, e) => { let step = e > s ? 1 : -1; return Array.from({ length: Math.abs(e - s) + 1 }, (_, i) => s + i * step); })(n - 1, 0));
        return ({ face, xyz, name: ngonal(n) + ` pyramid` });
    },
};
// Syntactic sugar to deal with weird TypeScript typing:
function æ(A, i) { return A.at(i); }
function wordsof(s) { return 1 + (s.match(/\s+/g)?.length ?? 0); }
const specialJoins = {
    Tetrahedron: 'Cube',
    Cube: 'Rhombic dodecahedron', Octahedron: 'Rhombic dodecahedron',
    Dodecahedron: 'Rhombic triacontahedron', Icosahedron: 'Rhombic triacontahedron',
    'Rhombic dodecahedron': 'Deltoidal icositetrahedron',
    'Rhombic triacontahedron': 'Deltoidal hexecontahedron',
    'pentakis Dodecahedron': 'Rhombic enneacontahedron',
};
const specialKis = {
    'Rhombic dodecahedron': 'Disdyakis dodecahedron',
    'Rhombic triacontahedron': 'Disdyakis triacontahedron',
};
const kisWords = { 3: 'triakis', 4: 'tetrakis', 5: 'pentakis', 6: 'hexakis', 7: 'heptakis' };
function kisjoin(P, notation, digits, join) {
    // kis and join are closely related operations. Both of them add a
    // pyramid on a selection of faces; join then further deletes any
    // _original_ edge bordered by two _new_ triangles, producing a quad.
    // Faces are selected by their numbers of sides, using the given digits.
    // If there are none, all faces are used. Otherwise, the digits are turned
    // into a list of numbers by breaking after every digit except as needed
    // to prevent leading 0s or isolated 1s or 2s (since no face has one or
    // two sides); this way you can list any subset of the numbers 3 - 32,
    // which is plenty.
    // The operation is then applied just to faces with the numbers of edges on
    // the list. e.g. k3412 will add pyramids to the triangles, quads, and
    // dodecagon faces.
    const allowed = parseSides(digits);
    // first collect a directory from face indices to new vertex numbers
    let nextVertex = P.xyz.length;
    const newVixes = (() => {
        const results3 = [];
        for (let ref = P.face, i5 = 0, len = ref.length; i5 < len; i5++) {
            const f = ref[i5];
            results3.push(!digits || allowed.includes(f.length) ? nextVertex++ : 0);
        }
        return results3;
    })();
    if (nextVertex === P.xyz.length)
        return P; // nothing to do
    const xyz = P.xyz.concat(faceCenters(P).filter((f, ix) => newVixes[ix]));
    const face = [];
    for (let ref1 = P.face, i6 = 0, len1 = ref1.length; i6 < len1; i6++) {
        const ix = i6;
        const f = ref1[i6];
        const v = newVixes[ix];
        if (v === 0) {
            face.push(f.slice());
            continue;
        }
        // Add the pyramid, possibly eliding edges:
        for (let i7 = 0, len2 = f.length; i7 < len2; i7++) {
            const jx = i7;
            const w = f[i7];
            const pw = æ(f, jx - 1);
            let neighbor = 0;
            if (join) {
                neighbor = P.face.findIndex((g, gx) => {
                    return gx !== ix && g.includes(w) && g.includes(pw);
                });
            }
            if (join && newVixes[neighbor]) { // elide this edge
                if (pw < w) { // avoid adding same face twice
                    face.push([v, pw, newVixes[neighbor], w]);
                }
            }
            else
                face.push([v, pw, w]);
        }
    }
    // Nomenclature
    let name;
    if (P.name && wordsof(P.name) < 3) { // don't go hog-wild with naming
        if (join) {
            if (!digits) {
                name = specialJoins[P.name] ?? 'joined ' + P.name;
            }
        }
        else {
            let size = P.face[0].length;
            if (!P.face.every((f) => f.length === size)) {
                size = 0;
            }
            if (!size && digits)
                size = Number(digits) || 0;
            // very special case
            if (size === 5 && P.name === 'pentagonal antiprism') {
                name = 'Icosahedron';
            }
            else if ((!digits || Number(digits) === size) && size in kisWords) {
                name = (specialKis[P.name]
                    ?? kisWords[size] + ' ' + P.name);
            }
        }
    }
    // Cheaty super special case
    if (notation === 'jk5djP5') {
        name = 'dual elongated pentagonal orthobirotunda';
    }
    // Done, fix up the vertices a bit
    return adjustXYZ({ face, xyz, name }, notation, 3);
}
const specialGyro = {
    Cube: 'pentagonal icositetrahedron', Octahedron: 'pentagonal icositetrahedron',
    Icosahedron: 'pentagonal hexecontahedron',
    Dodecahedron: 'pentagonal hexecontahedron',
};
// how enums ought to work?
const FromCenter = Symbol();
const AlongEdge = Symbol();
function gyropel(P, notation, digits, ...ways) {
    // gyro and propellor are closely related operations. Both of them add new
    // vertices one third of the way along each edge of each face selected
    // by the digits argument (see kisjoin). They then differ in what edges
    // are drawn to the new vertices. In gyro, another new vertex is added
    // at the center of each face and connected to each of them; in propellor,
    // they are just connected in sequence. For completeness, we also allow
    // both at the same time, which is equivalent to propellor followed by kis
    // just on the new rotated faces.
    const fromCenter = ways.includes(FromCenter);
    const alongEdge = ways.includes(AlongEdge);
    if (!(fromCenter || alongEdge))
        return P; // nothing to do
    const allowed = parseSides(digits);
    // first collect a directory from directed edges to new vertex numbers
    const xyz = P.xyz.slice();
    const startV = xyz.length;
    const edgeV = [];
    for (let ref2 = P.face, i8 = 0, len3 = ref2.length; i8 < len3; i8++) {
        const f = ref2[i8];
        if (digits && !allowed.includes(f.length))
            continue;
        for (let i9 = 0, len4 = f.length; i9 < len4; i9++) {
            const ix = i9;
            const v = f[i9];
            const pv = æ(f, ix - 1);
            (edgeV[pv] ??= [])[v] = xyz.length;
            xyz.push(lerp(xyz[pv], xyz[v], 1 / 3));
        }
    }
    if (xyz.length === startV)
        return P; // nothing to do
    // Now revisit each face, accumulating the new faces.
    const face = [];
    const centers = fromCenter ? faceCenters(P) : [];
    for (let ref3 = P.face, i10 = 0, len5 = ref3.length; i10 < len5; i10++) {
        const fx = i10;
        const f = ref3[i10];
        if (digits && !allowed.includes(f.length)) {
            // Just collect all of the vertices around
            const newFace = [];
            for (let i11 = 0, len6 = f.length; i11 < len6; i11++) {
                const ix = i11;
                const v = f[i11];
                const pv = æ(f, ix - 1);
                const reverseV = edgeV[v]?.[pv] ?? -1;
                if (reverseV >= 0)
                    newFace.push(reverseV);
                newFace.push(v);
            }
            face.push(newFace);
            continue;
        }
        const centerV = xyz.length;
        if (fromCenter)
            xyz.push(centers[fx]);
        let aroundOutside = [];
        for (let i12 = 0, len7 = f.length; i12 < len7; i12++) {
            const ix = i12;
            const v = f[i12];
            const pv = æ(f, ix - 1);
            const ppv = æ(f, ix - 2);
            const firstNew = edgeV[ppv][pv];
            const newSection = [firstNew];
            const reverseV = edgeV[pv][ppv] ?? -1;
            if (reverseV >= 0)
                newSection.push(reverseV);
            newSection.push(pv);
            const secondNew = edgeV[pv][v];
            if (alongEdge) {
                newSection.push(secondNew);
                face.push(newSection);
                if (fromCenter)
                    face.push(ð(centerV, firstNew, secondNew));
                else
                    aroundOutside.push(firstNew);
            }
            else if (fromCenter) {
                newSection.push(secondNew, centerV);
                face.push(newSection);
            }
            else
                concatAssign(aroundOutside, newSection);
        }
        if (aroundOutside.length)
            face.push(aroundOutside);
    }
    let name;
    let nw = 0;
    if ((xor(fromCenter, alongEdge))
        && P.name && !digits && (nw = wordsof(P.name)) < 3) {
        if (alongEdge) {
            if (nw === 1)
                name = 'propello' + P.name;
            else
                name = 'propellorized ' + P.name;
        }
        else {
            if (P.name in specialGyro)
                name = specialGyro[P.name];
            else if (nw === 1)
                name = 'gyro' + P.name;
            else
                name = 'gyrated ' + P.name;
        }
    }
    // Done, fix up the vertices a bit
    return adjustXYZ({ face, xyz, name }, notation, 3);
}
function parseSides(digits) {
    if (!digits)
        return [];
    const tooSmall = ['1', '2'];
    const last = digits.length - 1;
    const ret = [];
    let current = '';
    let pos = 0;
    while (pos <= last) {
        current += digits[pos++];
        const nextDigit = digits[pos];
        if (tooSmall.includes(current)
            || nextDigit === '0'
            || pos == last && tooSmall.includes(nextDigit)) {
            continue;
        }
        ret.push(parseInt(current));
        current = '';
    }
    return ret;
}
const transforms = {
    k: (P, notation, digits) => {
        return kisjoin(P, notation, digits, false);
    },
    j: (P, notation, digits) => {
        return kisjoin(P, notation, digits, true);
    },
    g: (P, notation, digits) => {
        return gyropel(P, notation, digits, FromCenter);
    },
    p: (P, notation, digits) => {
        return gyropel(P, notation, digits, AlongEdge);
    },
    f: (P, notation, digits) => {
        return gyropel(P, notation, digits, AlongEdge, FromCenter);
    },
    r: (P) => {
        return ({ face: (() => { const results4 = []; for (let ref4 = P.face, i13 = 0, len8 = ref4.length; i13 < len8; i13++) {
                const f = ref4[i13];
                results4.push(f.toReversed());
            } return results4; })(),
            xyz: (() => { const results5 = []; for (let ref5 = P.xyz, i14 = 0, len9 = ref5.length; i14 < len9; i14++) {
                const v = ref5[i14];
                results5.push(scale(copy(v), -1));
            } return results5; })(),
            name: P.name });
    }, // should we record that it's mirrored?
    d: (P, notation, digits) => {
        if (digits) {
            console.error(`Ignoring ${digits} arg of d in ${notation}`);
        }
        // Create a "fake" of P and adjust it (so we don't disturb the
        // cached P), which will create the dual
        const parentNotation = notation.slice(1 + digits.length);
        adjustXYZ({ face: P.face.slice(), xyz: P.xyz, name: P.name }, parentNotation, 1);
        return polyCache[`d${parentNotation}`];
    },
    c: (P, notation, digits) => {
        return ({ face: P.face.slice(),
            xyz: canonicalXYZ(P, notation, Number(digits) || 10),
            name: P.name });
    },
    x: approxCanonicalize,
}; // iterative direct adjustment algorithm
function dispatch(op, digits, P, mynotation) {
    // Note mynotation starts with op!
    let ret1 = P;
    if (op in seeds) {
        ret1 = seeds[op](Number(digits) || 0);
    }
    else if (op in transforms) {
        ret1 = transforms[op](P, mynotation, digits);
    }
    else {
        console.error(`Unknown operation ${op}${digits} in ${mynotation}.`);
        ret1 = polyCache.T;
    }
    polyCache[mynotation] = ret1;
    return ret1;
}
function topoDual(P) {
    // Note this maintains correspondence between V and F indices, but
    // places every vertex at the origin! Don't use without geometrizing
    // in some way.
    return ({ face: (() => {
            const results6 = [];
            for (let end1 = P.xyz.length, i15 = 0, asc4 = 0 <= end1; asc4 ? i15 < end1 : i15 > end1; asc4 ? ++i15 : --i15) {
                const v = i15;
                const infaces = // gather labeled list of faces contining v
                 (() => {
                    const results7 = [];
                    let i16 = 0;
                    for (const f of P.face) {
                        const index = i16++;
                        if (!f.includes(v))
                            continue;
                        results7.push(ð(f, index));
                    }
                    return results7;
                })();
                const start = infaces[0][1];
                let current = start;
                const newface = [];
                do {
                    const verts = P.face[current];
                    const preV = æ(verts, verts.indexOf(v) - 1);
                    const nextIx = infaces.findIndex(([face, label]) => {
                        return label !== current && face.includes(preV);
                    });
                    current = infaces[nextIx][1];
                    newface.push(current);
                    if (newface.length > infaces.length) {
                        console.error('In topoDual: Malformed polyhedron', P);
                        break;
                    }
                } while (!(current === start));
                results6.push(newface);
            }
            return results6;
        })(),
        xyz: Array(P.face.length).fill([0, 0, 0]), // warning, every vertex is ===
        name: dualName(P.name) });
}
function geomDual(P) {
    const ret2 = topoDual(P);
    ret2.xyz = approxDualVertices(P);
    return ret2;
}
function approxDualVertices(P) {
    return P.face.map((f) => approxDualVertex(f, P.xyz));
}
function dot(v, w) {
    return v.reduce((l, r, i) => l + r * w[i], 0);
}
function approxDualVertex(f, v) {
    // For each edge of f, there is a plane containing it perpendicular
    // to the line joining the origin to its nearest approach to the origin.
    // This function returns the point closest to being on all of those planes
    // (in the least-squares sense).
    // This method seems to work OK when the neighborhood of f is convex,
    // and very poorly otherwise. So it probably would not provide any better
    // canonicalization than other methods of approximating the dual.
    // It might be better to replace this with the approximate intersection of the
    // reciprocal planes to each of the vertices of the face...
    const normals = (() => { const results8 = []; for (let end2 = f.length, i17 = 0, asc5 = 0 <= end2; asc5 ? i17 < end2 : i17 > end2; asc5 ? ++i17 : --i17) {
        const i = i17;
        results8.push(tangentPoint(v[æ(f, i - 1)], v[f[i]]));
    } return results8; })();
    const sqlens = normals.map(mag2);
    const columns = (() => { const results9 = []; for (let i18 = 0; i18 <= 2; ++i18) {
        const i = i18;
        results9.push(normals.map($1 => $1[i]));
    } return results9; })();
    const target = (() => { const results10 = []; for (let i19 = 0; i19 <= 2; ++i19) {
        const i = i19;
        results10.push(dot(columns[i], sqlens));
    } return results10; })();
    const CMsource = (() => {
        const results11 = [];
        for (let i20 = 0; i20 <= 2; ++i20) {
            const c = i20;
            results11.push((() => { const results12 = []; for (let i21 = 0; i21 <= 2; ++i21) {
                const r = i21;
                results12.push(dot(columns[r], columns[c]));
            } return results12; })());
        }
        return results11;
    })();
    const cramerD = det(...CMsource);
    if (Math.abs(cramerD) < 1e-6) {
        console.error(`Face ${f} of ${v.map((p) => '[' + p + ']')} ill conditioned`);
        return [0, 0, 0];
    }
    return [det(target, CMsource[1], CMsource[2]) / cramerD,
        det(CMsource[0], target, CMsource[2]) / cramerD,
        det(CMsource[0], CMsource[1], target) / cramerD];
}
function det(a, b, c) {
    return a[0] * b[1] * c[2] + a[1] * b[2] * c[0] + a[2] * b[0] * c[1]
        - a[2] * b[1] * c[0] - a[1] * b[0] * c[2] - a[0] * b[2] * c[1];
}
function sub(a, b) {
    return diminish(copy(a), b);
}
function tangentPoint(v, w) {
    const d = sub(w, v);
    return sub(v, scale(d, dot(d, v) / mag2(d)));
}
function faceCenters(P) {
    const results13 = [];
    for (let ref6 = P.face, i22 = 0, len10 = ref6.length; i22 < len10; i22++) {
        const face = ref6[i22];
        results13.push(scale(face.reduce((ctr, v) => accumulate(ctr, P.xyz[v]), [0, 0, 0]), 1 / face.length));
    }
    ;
    return results13;
}
function dualName(p) {
    if (!('Octahedron' in specialDuals)) {
        // one-time reversal of all special duals
        for (const poly in specialDuals) {
            const dual = specialDuals[poly];
            specialDuals[dual] = poly;
        }
    }
    if (!p)
        return undefined;
    if (p in specialDuals) {
        return specialDuals[p];
    }
    const words = p.split(' ');
    if (words[0] === 'dual')
        return words.slice(1).join(' ');
    if (words.length === 2) {
        let m1;
        if (m1 = words[1], m1 === 'prism') {
            return words[0] + ' bipyramid';
        }
        else if (m1 === 'bipyramid') {
            return words[0] + ' prism';
        }
        else if (m1 === 'antiprism') {
            return words[0] + ' trapezohedron';
        }
        else if (m1 === 'trapezohedron') {
            return words[0] + ' antiprism';
        }
        else if (m1 === 'pyramid') {
            return p;
        }
    } // self-dual
    return 'dual ' + p;
}
function adjustXYZ(P, notation, iterations = 1) {
    const dualNotation = 'd' + notation;
    let D = topoDual(P);
    if (dualNotation in polyCache) {
        console.error('Creating', notation, '_after_ its dual');
        D = polyCache[dualNotation];
    }
    for (let i23 = 1, asc6 = 1 <= iterations; asc6 ? i23 <= iterations : i23 >= iterations; asc6 ? ++i23 : --i23) {
        const iter = i23;
        D.xyz = reciprocalC(P);
        P.xyz = reciprocalC(D);
    }
    polyCache[dualNotation] = D;
    return P;
}
function reciprocalC(P) {
    const ret3 = faceCenters(P);
    for (let ref7 = ret3, i24 = 0, len11 = ref7.length; i24 < len11; i24++) {
        const v = ref7[i24];
        scale(v, 1 / mag2(v));
    }
    return ret3;
}
function canonicalXYZ(P, notation, iterations) {
    const dualNotation = 'd' + notation;
    let D = topoDual(P);
    if (dualNotation in polyCache) {
        console.error('Creating', notation, '_after_ its dual');
        D = polyCache[dualNotation];
    }
    const tempP = Object.assign({}, P); // algorithm is read-only on original data
    if (iterations < 1)
        iterations = 1;
    for (let i25 = 1, asc7 = 1 <= iterations; asc7 ? i25 <= iterations : i25 >= iterations; asc7 ? ++i25 : --i25) {
        const iter = i25;
        D.xyz = reciprocalN(tempP);
        tempP.xyz = reciprocalN(D);
    }
    polyCache[dualNotation] = D;
    return tempP.xyz;
}
function cross(v, w) {
    return [v[1] * w[2] - v[2] * w[1], v[2] * w[0] - v[0] * w[2], v[0] * w[1] - v[1] * w[0]];
}
function reciprocalN(P) {
    const results14 = [];
    for (let ref8 = P.face, i26 = 0, len12 = ref8.length; i26 < len12; i26++) {
        const f = ref8[i26];
        const centroid = ð(0, 0, 0);
        const normal = ð(0, 0, 0);
        let meanEdgeRadius = 0;
        for (let i27 = 0, len13 = f.length; i27 < len13; i27++) {
            const ix = i27;
            const vlabel = f[i27];
            const v = P.xyz[vlabel];
            accumulate(centroid, v);
            const pv = P.xyz[æ(f, ix - 1)];
            const ppv = P.xyz[æ(f, ix - 2)];
            // original doc says unit normal but below isn't. Didn't help to try, tho
            const nextNormal = cross((sub(pv, ppv)), (sub(v, pv)));
            // or instead, just chop down big ones: (didn't work either)
            // magNext := mag nextNormal
            // if magNext > 1 then scale nextNormal, 1/magNext
            accumulate(normal, nextNormal);
            meanEdgeRadius += mag(tangentPoint(pv, v));
        }
        scale(centroid, 1 / f.length);
        scale(normal, 1 / mag(normal));
        meanEdgeRadius /= f.length;
        scale(normal, dot(centroid, normal));
        scale(normal, 1 / mag2(normal)); // invert in unit sphere
        results14.push(scale(normal, (1 + meanEdgeRadius) / 2));
    }
    ;
    return results14;
}
function dist(a, b) { return mag(sub(a, b)); }
// adapted from Polyhedronisme https://levskaya.github.io/polyhedronisme/
function approxCanonicalize(P, notation, digits) {
    const THRESHOLD = 1e-6;
    // A difficulty is that the planarizing sometimes has the effect of
    // "folding over" neighboring faces, in which case all is lost.
    // Keeping the weight of the edge smoothing high compared to planarizing
    // seems to help with that.
    const EDGE_SMOOTH_FACTOR = 0.5;
    const PLANARIZE_FACTOR = 0.1;
    const edge = edges(P);
    const xyz = P.xyz.map(copy);
    const V = xyz.length;
    normalizeEdges(xyz, edge);
    for (let end3 = Number(digits) || 10, i28 = 1, asc8 = 1 <= end3; asc8 ? i28 <= end3 : i28 >= end3; asc8 ? ++i28 : --i28) {
        const iter = i28;
        const start = xyz.map(copy);
        smoothEdgeDists(xyz, edge, EDGE_SMOOTH_FACTOR);
        normalizeEdges(xyz, edge);
        planarize(xyz, P.face, PLANARIZE_FACTOR);
        normalizeEdges(xyz, edge);
        if (Math.max(...(() => { const results15 = []; for (let i29 = 0, asc9 = 0 <= V; asc9 ? i29 < V : i29 > V; asc9 ? ++i29 : --i29) {
            const i = i29;
            results15.push(dist(xyz[i], start[i]));
        } return results15; })()) < THRESHOLD) {
            break;
        }
    }
    return ({ face: P.face.slice(), xyz, name: P.name });
}
function edges(P) {
    const ret4 = [];
    for (let ref9 = P.face, i30 = 0, len14 = ref9.length; i30 < len14; i30++) {
        const f = ref9[i30];
        for (let i31 = 0, len15 = f.length; i31 < len15; i31++) {
            const ix = i31;
            const v = f[i31];
            const pv = æ(f, ix - 1);
            if (pv < v)
                ret4.push(ð(pv, v));
        }
    }
    return ret4;
}
function normalizeEdges(xyz, edge) {
    // Adjusts xyz so that edge tangentpoints have centroid at origin and
    // mean radius 1
    let edgeP = edge.map(([a, b]) => tangentPoint(xyz[a], xyz[b]));
    const edgeCentroid = centroid(edgeP);
    xyz.forEach((pt) => diminish(pt, edgeCentroid));
    const edgeScale = 1 / (mean(edge.map(([a, b]) => mag(tangentPoint(xyz[a], xyz[b])))));
    xyz.forEach((pt) => scale(pt, edgeScale));
}
function centroid(xyz) {
    return scale(xyz.reduce(accumulate, ð(0, 0, 0)), 1 / xyz.length);
}
function smoothEdgeDists(xyz, edge, fudge) {
    // Attempts in the most straightforward way possible to reduce the
    // variance of the radii of the edgepoints
    const V = xyz.length;
    const adj = (() => { const results16 = []; for (let i32 = 1, asc10 = 1 <= V; asc10 ? i32 <= V : i32 >= V; asc10 ? ++i32 : --i32) {
        const i = i32;
        results16.push(ð(0, 0, 0));
    } return results16; })();
    const edgeDistsStart = edge.map(([a, b]) => mag(tangentPoint(xyz[a], xyz[b])));
    for (let i33 = 0, len16 = edge.length; i33 < len16; i33++) {
        const [a, b] = edge[i33];
        const t = tangentPoint(xyz[a], xyz[b]);
        scale(t, (1 - mag(t)) / 2);
        accumulate(adj[a], t);
        accumulate(adj[b], t);
    }
    for (let i34 = 0, asc11 = 0 <= V; asc11 ? i34 < V : i34 > V; asc11 ? ++i34 : --i34) {
        const i = i34;
        accumulate(xyz[i], scale(adj[i], fudge));
    }
    const edgeDistsEnd = edge.map(([a, b]) => mag(tangentPoint(xyz[a], xyz[b])));
}
function summary(data) {
    return [Math.min(...data), Math.max(...data), mean(data),
        mean(data.map((x) => Math.abs(1 - x)))];
}
function planarize(xyz, face, fudge) {
    const V = xyz.length;
    const adj = (() => { const results17 = []; for (let i35 = 1, asc12 = 1 <= V; asc12 ? i35 <= V : i35 >= V; asc12 ? ++i35 : --i35) {
        const i = i35;
        results17.push(ð(0, 0, 0));
    } return results17; })();
    for (let i36 = 0, len17 = face.length; i36 < len17; i36++) {
        const f = face[i36];
        if (f.length === 3)
            continue; // triangles always planar
        const fxyz = (() => { const results18 = []; for (let i37 = 0, len18 = f.length; i37 < len18; i37++) {
            const v = f[i37];
            results18.push(xyz[v]);
        } return results18; })();
        const c = centroid(fxyz);
        const n = meanNormal(fxyz);
        if (dot(c, n) < 0)
            scale(n, -1);
        for (let i38 = 0, len19 = f.length; i38 < len19; i38++) {
            const v = f[i38];
            accumulate(adj[v], scale(copy(n), dot(n, (sub(c, xyz[v])))));
        }
    }
    for (let i39 = 0, asc13 = 0 <= V; asc13 ? i39 < V : i39 > V; asc13 ? ++i39 : --i39) {
        const i = i39;
        accumulate(xyz[i], scale(adj[i], fudge));
    }
}
function meanNormal(xyz) {
    const mNormal = ð(0, 0, 0);
    let [v1, v2] = xyz.slice(-2);
    for (let i40 = 0, len20 = xyz.length; i40 < len20; i40++) {
        const v3 = xyz[i40];
        const nextNormal = cross((sub(v2, v1)), (sub(v3, v2)));
        const magNext = mag(nextNormal);
        // reduce influence of long edges? (didn't seem to help in brief testing)
        // if magNext > 1 then scale nextNormal, 1/Math.sqrt magNext
        accumulate(mNormal, nextNormal);
        [v1, v2] = [v2, v3];
    } // shift over one
    return scale(mNormal, 1 / mag(mNormal));
}
function mean(a) {
    let m = 0;
    for (let i41 = 0, len21 = a.length; i41 < len21; i41++) {
        const e = a[i41];
        m += e;
    }
    return m / a.length;
}
// arithmetic on 3-vectors
function accumulate(basket, egg) {
    basket[0] += egg[0];
    basket[1] += egg[1];
    basket[2] += egg[2];
    return basket;
}
function diminish(basket, egg) {
    basket[0] -= egg[0];
    basket[1] -= egg[1];
    basket[2] -= egg[2];
    return basket;
}
function copy(a) {
    return ð(a[0], a[1], a[2]);
}
function mag2(a) {
    return a[0] * a[0] + a[1] * a[1] + a[2] * a[2];
}
function mag(a) {
    return Math.sqrt(mag2(a));
}
function normalize(v) {
    return scale(v, 1 / mag(v));
}
function unit() { }
function scale(subject, by) {
    subject[0] *= by;
    subject[1] *= by;
    subject[2] *= by;
    return subject;
}
function lerp(start, end, howFar) {
    const basket = scale(copy(start), 1 - howFar);
    return accumulate(basket, scale(copy(end), howFar));
}
// Feedback
function inform(x) {
    $('input[name="inform"]').val(x);
    return x;
}
// VRML97 generation
function outputVRML(notation, P) {
    const shortDescrip = P.name || `A ${P.face.length}-hedron`;
    return `#VRML V2.0 utf8
Group { children [
   WorldInfo { # Generated by GTW's reimplementation of GWH's Conway script.
      title "${notation}   ${stats(P)}"
      info "Generated by GTW's Conway-notation script inspired by GWH's.
By using this script, you agree that this image is released
into the public domain, although you are requested to cite
George Hart's Encyclopedia of Polyhedra as the source." }
   Background {
      groundColor [ .2 .5 1 ] # light blue
      skyColor    [ .2 .5 1 ] }
   NavigationInfo { type [ "EXAMINE" ] }
   DirectionalLight {direction -.5 -1  1 intensity 0.75}
   DirectionalLight {direction  .5  1 -1 intensity 0.75}
   Viewpoint { position 0 0 4.5 description "${shortDescrip}" }
   ${polyVRML(P, colorScheme())}
   Shape {
      appearance Appearance {
         material Material {
            diffuseColor 0 0 0 } }
      geometry IndexedLineSet {
         coord ${useVerts}
         coordIndex [
            ${edgeIndices(P)} ] } }
   ${showDual() ? polyVRML(geomDual(P), '0.5 0.5 0.5') : ''} ] }`;
}
function stats(P) {
    const edges = P.face.reduce((e, f) => e + f.length, 0) / 2;
    return `[${P.face.length} faces, ${edges} edges, ${P.xyz.length} vertices]`;
}
const useVerts = 'USE verts';
function defVerts(v) {
    return `DEF verts Coordinate {
         point [
            ${v.map($2 => $2.join(' ')).join(",\n            ")} ] }`;
}
function polyVRML(P, color) {
    const facePartition = color ? [P.face]
        : P.face.reduce(((parts, f) => {
            (parts[f.length] ??= []).push(f);
            return parts;
        }), []);
    let emittedCoords = '';
    const shapes = (() => {
        const results19 = [];
        for (const part of facePartition) {
            if (!part)
                continue;
            results19.push(`
   Shape {
      appearance Appearance {
         material Material {
            diffuseColor ${color || colorBySides(part[0].length)} } }
      geometry IndexedFaceSet {
         ccw FALSE
         coord ${emittedCoords ? useVerts : (emittedCoords = defVerts(P.xyz))}
         coordIndex [
            ${part.map($3 => $3.join(', ')).join(", -1,\n            ")}, -1 ] }}`);
        }
        return results19;
    })();
    return shapes.join("\n");
}
function colorScheme() {
    const button = document.getElementsByName('color')[0];
    return button.checked ? '1 1 1' : '';
}
function showDual() {
    return false;
}
const faceColors = {
    3: '0.9 0.3 0.3', // red
    4: '0.4 0.4 1.0', // blue
    5: '0.2 0.9 0.3', // green
    6: '0.9 0.9 0.2', // yellow
    7: '0.5 0.25 0.25', // brown
    8: '0.8 0.2 0.8', // magenta
    9: '0.5 0.2 0.8', // purple
    10: '0.1 0.9 0.9', // cyan
    12: '1.0 0.6 0.1',
}; // orange
function colorBySides(n) {
    if (n in faceColors) {
        return faceColors[n];
    }
    return '0.5 0.5 0.5';
} // gray
function filtmap(A, m) {
    return A.map(m).filter((e) => !!e);
}
function edgeIndices(P) {
    const sep = ",\n            ";
    return filtmap(P.face, (thisf) => {
        return filtmap(thisf, (v, ix, f) => {
            const preV = æ(f, ix - 1);
            return preV < v ? `${preV}, ${v}, -1` : '';
        }).join(sep);
    })
        .join(sep);
}
