function gcdpos(a, b) {
    while (b >= 1) {
        [a, b] = [b, a % b];
    }
    return a;
}
const shapeNamesRaw = `prism|dipyramid|compound: prism + dipyramid
antiprism|trapezohedron|compound: antiprism + trapezohedron
crossed antiprism|concave trapezohedron
compound: crossed antiprism + concave trapezohedron|polygon`;
const shapeName = shapeNamesRaw.split(/[|\n]/);
export function generateVRML(n, d, chk) {
    if (!chk.slice(6, 1 + 8 || 1 / 0).every($ => !$ || 2 < n / d && n / d < 3)) {
        return { err: 'Crossed antiprisms require 2 < numerator/denominator < 3, but '
                + `it is ${n}/${d} = ${n / d}` };
    }
    if (!(n > 0 && d > 0))
        return { err: 'Numerator and denominator should be positive' };
    const g = gcdpos(n, d);
    n /= g;
    d /= g;
    if (n < 3) {
        return {
            err: `Numerator in lowest terms (currently ${n}) needs to be at least 3.`,
        };
    }
    if (!(1 <= d && d < n)) {
        return { err: 'Denominator should be strictly between 1 and numerator.' };
    }
    if (d > n / 2)
        d = n - d; // equivalent
    const which = chk.findIndex($1 => $1);
    chk[which] = false;
    if (chk.some($2 => $2))
        return { err: 'Currently only one shape option may be checked.' };
    let name = d > 1 ? `${n}/${d}-gonal` :
        (() => {
            switch (n) {
                case 3: {
                    return 'triangular';
                }
                case 4: {
                    return 'square';
                }
                case 5: {
                    return 'pentagonal';
                }
                case 6: {
                    return 'hexagonal';
                }
                case 7: {
                    return 'heptagonal';
                }
                case 8: {
                    return 'octagonal';
                }
                default: {
                    return `${n}-gonal`;
                }
            }
        })();
    name += ' ' + shapeName[which];
    return ({ vrml: header(name) + vrmlBody(which, n, d) });
}
function header(name) {
    return `#VRML V2.0 utf8
WorldInfo {
   title  "${name}."
   info  [
      "Generated by GTW's reimplementation of GWH's prism generator"
      "By using this script, you agree that this image is released"
      "into the public domain, although you are requested to cite"
      "George Hart's Encyclopedia of Polyhedra as the source." ] }
Background {
   groundColor [ 0.2 0.5 0.9 ]
   skyColor [ 0.2 0.5 0.9 ] }
NavigationInfo { type [ "EXAMINE" ] }
DirectionalLight { direction -.5 -1 1 intensity 0.75 }
DirectionalLight { direction .5 1 -1 intensity 0.75 }
DEF General Viewpoint {
   position 6 0 6 orientation 0 1 0 0.78 description "General" }
DEF Face Viewpoint {
   position 0 -6 0 orientation 1 0 0 1.57 description "Face" }
DEF Top Viewpoint { position 0 0 8 description "Top" }
DEF Edge Viewpoint {
   position 7 0 0 orientation 0 1 0 1.57 description "Edge" }
DEF Inside Viewpoint {
   position 0 0 0 orientation 1 0 0 -1.57 description "Inside" }
DEF Far Viewpoint {
   position -25 0 25 orientation 0 1 0 -0.78 description "Far" }`;
}
// Useful constants
const tau = 2 * Math.PI;
// Array "postscript" operator
function ps(main, extra) {
    main.push(extra);
    return main;
}
// Array "preface" operator
function pf(extra, main) {
    main.unshift(extra);
    return main;
}
function shIx(name) {
    return shapeName.findIndex((sn) => sn.includes(name));
}
const shapeMethod = [];
const shapeMethods = {
    prism: (n, theta) => {
        const name = 'Prism';
        const h = Math.sin(theta / 2);
        const coords = ps(ps(orb(1, n, h, theta).concat(orb(1, n, -h, theta)), [0, 0, h]), [0, 0, -h]);
        const top = (() => { const results = []; for (let i1 = 0, asc = 0 <= n; asc ? i1 < n : i1 > n; asc ? ++i1 : --i1) {
            const i = i1;
            results.push([i, (i + 1) % n, 2 * n]);
        } return results; })();
        const bottom = (() => { const results1 = []; for (let i2 = 0, asc1 = 0 <= n; asc1 ? i2 < n : i2 > n; asc1 ? ++i2 : --i2) {
            const i = i2;
            results1.push([(i + 1) % n + n, i + n, 2 * n + 1]);
        } return results1; })();
        const vrml = firstFaces(name, '0.8 0.7 0.5', coords, top.concat(bottom));
        const sides = (() => { const results2 = []; for (let i3 = 0, asc2 = 0 <= n; asc2 ? i3 < n : i3 > n; asc2 ? ++i3 : --i3) {
            const i = i3;
            results2.push([i, (i + 1) % n, (i + 1) % n + n, i + n]);
        } return results2; })();
        laterFaces(vrml, name, '0.3 0.5 0.9', sides);
        laterLines(vrml, name, sides);
        return emitObject(vrml);
    },
    dipyr: (n, theta) => {
        const name = 'Dipyramid';
        const c = 1 / Math.cos(theta / 2);
        const d = 1 / Math.sin(theta / 2);
        const coords = ps(pf([0, 0, d], orb(c, n, 0, theta, 0.5)), [0, 0, -d]);
        const cap = (() => { const results3 = []; for (let i4 = 1, asc3 = 1 <= n; asc3 ? i4 <= n : i4 >= n; asc3 ? ++i4 : --i4) {
            const i = i4;
            results3.push([0, i, i % n + 1]);
        } return results3; })();
        const cup = (() => { const results4 = []; for (let i5 = 1, asc4 = 1 <= n; asc4 ? i5 <= n : i5 >= n; asc4 ? ++i5 : --i5) {
            const i = i5;
            results4.push([i % n + 1, i, n + 1]);
        } return results4; })();
        const vrml = firstFaces(name, '0.9 0.3 0.3', coords, cap.concat(cup));
        const edges = pf(((s, e) => { let step = e > s ? 1 : -1; return Array.from({ length: Math.abs(e - s) + 1 }, (_, i) => s + i * step); })(1, n), (() => { const results5 = []; for (let i6 = 1, asc5 = 1 <= n; asc5 ? i6 <= n : i6 >= n; asc5 ? ++i6 : --i6) {
            const i = i6;
            results5.push([0, i, n + 1]);
        } return results5; })());
        laterLines(vrml, name, edges);
        return emitObject(vrml);
    },
    compound: (n, theta) => {
        return shapeMethod[shIx('prism')](n, theta) + shapeMethod[shIx('dipyr')](n, theta);
    },
    antip: (n, theta) => {
        const name = 'Antiprism';
        const { h, r } = antiprismDims(theta);
        const coords = (ps(ps(orb(r, n, h, theta).concat(orb(r, n, -h, theta, 0.5)), [0, 0, h]), [0, 0, -h]));
        const top = (() => { const results6 = []; for (let i7 = 0, asc6 = 0 <= n; asc6 ? i7 < n : i7 > n; asc6 ? ++i7 : --i7) {
            const i = i7;
            results6.push([i, 2 * n, (i + 1) % n]);
        } return results6; })();
        const bottom = (() => { const results7 = []; for (let i8 = 0, asc7 = 0 <= n; asc7 ? i8 < n : i8 > n; asc7 ? ++i8 : --i8) {
            const i = i8;
            results7.push([(i + 1) % n + n, 2 * n + 1, i + n]);
        } return results7; })();
        const vrml = firstFaces(name, '0.8 0.7 0.5', coords, top.concat(bottom));
        const topsides = (() => { const results8 = []; for (let i9 = 0, asc8 = 0 <= n; asc8 ? i9 < n : i9 > n; asc8 ? ++i9 : --i9) {
            const i = i9;
            results8.push([(i + 1) % n, i + n, i]);
        } return results8; })();
        const botsides = (() => { const results9 = []; for (let i10 = 0, asc9 = 0 <= n; asc9 ? i10 < n : i10 > n; asc9 ? ++i10 : --i10) {
            const i = i10;
            results9.push([i, i + n, (n + i - 1) % n + n]);
        } return results9; })();
        laterFaces(vrml, name, '0.9 0.3 0.4', topsides.concat(botsides));
        laterLines(vrml, name, [ps(((s, e) => { let step = e > s ? 1 : -1; return Array.from({ length: Math.abs(e - s) }, (_, i) => s + i * step); })(0, n), 0), ps(((s, e) => { let step = e > s ? 1 : -1; return Array.from({ length: Math.abs(e - s) }, (_, i) => s + i * step); })(n, 2 * n), n), ...topsides]);
        return emitObject(vrml);
    },
    trap: (n, theta) => {
        const name = 'Trapezohedron';
        const { h, r } = antiprismDims(theta);
        const [t, z] = dualPoint(r * Math.cos(theta / 2), h, r, -h);
        const botzig = ps(orb(t, n, -z, theta), [0, 0, -1 / h]);
        const topzig = ps(orb(t, n, z, theta, 0.5), [0, 0, 1 / h]);
        const coords = alternate(botzig, topzig);
        const faces = (() => {
            const results10 = [];
            for (let end = 2 * n, i11 = 0, asc10 = 0 <= end; asc10 ? i11 < end : i11 > end; asc10 ? ++i11 : --i11) {
                const i = i11;
                if (i % 2)
                    results10.push([2 * n + 1, (i + 2) % (2 * n), (i + 1) % (2 * n), i]);
                else
                    results10.push([2 * n, i, i + 1, (i + 2) % (2 * n)]);
            }
            return results10;
        })();
        const vrml = firstFaces(name, '0.3 0.5 0.9', coords, faces);
        const edges = (() => {
            const results11 = [];
            for (let end1 = 2 * n, i12 = 0, asc11 = 0 <= end1; asc11 ? i12 < end1 : i12 > end1; asc11 ? ++i12 : --i12) {
                const i = i12;
                results11.push(i % 2 ? [2 * n + 1, i, (i + 1) % (2 * n)] : [2 * n, i, i + 1]);
            }
            return results11;
        })();
        laterLines(vrml, name, edges);
        return emitObject(vrml);
    },
    'antiprism + trap': (n, theta) => {
        return shapeMethod[shIx('antip')](n, theta) + shapeMethod[shIx('trap')](n, theta);
    },
    crossed: (n, theta) => {
        const name = 'CrossedAntiprism';
        const { h, r } = crossedDims(theta);
        const coords = (ps(ps(orb(r, n, h, theta).concat(orb(-r, n, -h, theta, 0.5)), [0, 0, h]), [0, 0, -h]));
        const top = (() => { const results12 = []; for (let i13 = 0, asc12 = 0 <= n; asc12 ? i13 < n : i13 > n; asc12 ? ++i13 : --i13) {
            const i = i13;
            results12.push([i, (i + 1) % n, 2 * n]);
        } return results12; })();
        const bottom = (() => { const results13 = []; for (let i14 = 0, asc13 = 0 <= n; asc13 ? i14 < n : i14 > n; asc13 ? ++i14 : --i14) {
            const i = i14;
            results13.push([(i + 1) % n + n, i + n, 2 * n + 1]);
        } return results13; })();
        const vrml = firstFaces(name, '0.8 0.7 0.5', coords, top.concat(bottom));
        const topsides = (() => { const results14 = []; for (let i15 = 0, asc14 = 0 <= n; asc14 ? i15 < n : i15 > n; asc14 ? ++i15 : --i15) {
            const i = i15;
            results14.push([(i + 1) % n, i + n, i]);
        } return results14; })();
        const botsides = (() => { const results15 = []; for (let i16 = 0, asc15 = 0 <= n; asc15 ? i16 < n : i16 > n; asc15 ? ++i16 : --i16) {
            const i = i16;
            results15.push([i, i + n, (n + i - 1) % n + n]);
        } return results15; })();
        laterFaces(vrml, name, '0.9 0.3 0.4', topsides.concat(botsides));
        laterLines(vrml, name, [ps(((s, e) => { let step = e > s ? 1 : -1; return Array.from({ length: Math.abs(e - s) }, (_, i) => s + i * step); })(0, n), 0), ps(((s, e) => { let step = e > s ? 1 : -1; return Array.from({ length: Math.abs(e - s) }, (_, i) => s + i * step); })(n, 2 * n), n), ...topsides]);
        return emitObject(vrml);
    },
    concave: (n, theta, d) => {
        const name = 'ConcaveTrapezohedron';
        const { h, r } = crossedDims(theta);
        const [t, z] = dualPoint(r * Math.cos(theta / 2), h, -r, -h);
        d ??= Math.floor((n - 1) / 2); // won't be needed, for TypeScript's sake
        const eta = tau * (n - d) / n;
        const botzig = ps(orb(-t, n, -z, eta), [0, 0, -1 / h]);
        const topzig = ps(orb(-t, n, z, eta, 0.5), [0, 0, 1 / h]);
        const coords = alternate(botzig, topzig);
        const halfFaces = (() => {
            const results16 = [];
            for (let end2 = 2 * n, i17 = 0, asc16 = 0 <= end2; asc16 ? i17 < end2 : i17 > end2; asc16 ? ++i17 : --i17) {
                const i = i17;
                results16.push(i % 2 ? [(i + 1) % (2 * n), i, 2 * n + 1] : [2 * n, i, i + 1]);
            }
            return results16;
        })();
        const otherFaces = (() => {
            const results17 = [];
            for (let end3 = 2 * n, i18 = 0, asc17 = 0 <= end3; asc17 ? i18 < end3 : i18 > end3; asc17 ? ++i18 : --i18) {
                const i = i18;
                results17.push(i % 2 ? [2 * n + 1, (i + 2) % (2 * n), (i + 1) % (2 * n)] : [2 * n, i + 1, (i + 2) % (2 * n)]);
            }
            return results17;
        })();
        const vrml = firstFaces(name, '0.3 0.5 0.9', coords, halfFaces.concat(otherFaces));
        laterLines(vrml, name, halfFaces);
        return emitObject(vrml);
    },
    'prism + concave': (n, theta, d) => {
        return shapeMethod[shIx('cross')](n, theta) + shapeMethod[shIx('conc')](n, theta, d);
    },
    polygon: (n, theta) => {
        const name = 'Polygon';
        const coords = ps(orb(1, n, 0, theta), [0, 0, 0]);
        const faces = (() => { const results18 = []; for (let i19 = 0, asc18 = 0 <= n; asc18 ? i19 < n : i19 > n; asc18 ? ++i19 : --i19) {
            const i = i19;
            results18.push([i, (i + 1) % n, n]);
        } return results18; })();
        const vrml = firstFaces(name, '0.5 0.8 0.5', coords, faces);
        laterLines(vrml, name, [ps(((s, e) => { let step = e > s ? 1 : -1; return Array.from({ length: Math.abs(e - s) }, (_, i) => s + i * step); })(0, n), 0)]);
        return emitObject(vrml);
    },
};
for (const key in shapeMethods) {
    const method = shapeMethods[key];
    shapeMethod[shIx(key)] = method;
}
function antiprismDims(theta) {
    let h = Math.sqrt(1 - 4 / (4 + 2 * Math.cos(theta / 2) - 2 * Math.cos(theta)));
    let r = Math.sqrt(1 - h * h);
    const f = Math.sqrt(h * h + (r * Math.cos(theta / 2)) ** 2);
    h /= f;
    r /= f;
    return ({ h, r });
}
function crossedDims(theta) {
    let r = 2 / Math.sqrt(4 - 2 * Math.cos(theta / 2) - 2 * Math.cos(theta));
    let h = Math.sqrt(1 - r * r);
    let f = Math.sqrt(h * h + (r * Math.cos(theta / 2)) ** 2);
    r /= f;
    h /= f;
    return ({ h, r });
}
function dualPoint(x1, y1, x2, y2) {
    const len = Math.sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));
    const vx = (y2 - y1) / len;
    const vy = (x1 - x2) / len;
    const d = vx * x1 + vy * y1;
    return [vx / d, vy / d];
}
function vrmlBody(which, n, d) {
    const theta = tau * d / n;
    return shapeMethod[which](n, theta, d);
}
function orb(r, n, height, theta, t = 0) {
    const rho = t * theta;
    const results19 = [];
    for (let i20 = 0, asc19 = 0 <= n; asc19 ? i20 < n : i20 > n; asc19 ? ++i20 : --i20) {
        const i = i20;
        results19.push([r * Math.cos(rho + i * theta), r * Math.sin(rho + i * theta), height]);
    }
    ;
    return results19;
}
function alternate(a, b) {
    const ret = [];
    for (let i21 = 0, len1 = a.length; i21 < len1; i21++) {
        const i = i21;
        const n = a[i21];
        ret.push(n, b[i]);
    }
    return ret;
}
function firstFaces(name, color, coords, faces) {
    const ret1 = [];
    startShape(ret1, color);
    ret1.push('geometry IndexedFaceSet {', `coord DEF ${name}Coords Coordinate { point [`);
    for (let i22 = 0, len2 = coords.length; i22 < len2; i22++) {
        const coord = coords[i22];
        ret1.push(coord.join(' ') + ',');
    }
    ret1.push('] }', 'creaseAngle 0 solid FALSE coordIndex [');
    for (let i23 = 0, len3 = faces.length; i23 < len3; i23++) {
        const face = faces[i23];
        ret1.push(face.join(', ') + ', -1,');
    }
    ret1.push('] } }');
    return ret1;
}
function startShape(container, color, colType = 'diffuse') {
    container.push('Shape { appearance Appearance {', '  material Material {', `     ${colType}Color ${color} } }`);
}
function laterFaces(container, name, color, faces) {
    startShape(container, color);
    container.push('geometry IndexedFaceSet {', `coord USE ${name}Coords`, 'creaseAngle 0 solid FALSE coordIndex [');
    for (let i24 = 0, len4 = faces.length; i24 < len4; i24++) {
        const face = faces[i24];
        container.push(face.join(', ') + ', -1,');
    }
    container.push('] } }');
}
function laterLines(container, name, edges) {
    startShape(container, '0 0 0', 'emissive'); // per VRML97 standard
    container.push('geometry IndexedLineSet {', `coord USE ${name}Coords`, 'coordIndex [');
    for (let i25 = 0, len5 = edges.length; i25 < len5; i25++) {
        const edge = edges[i25];
        container.push(edge.join(', ') + ', -1,');
    }
    container.push('] } }');
}
function emitObject(container) { return container.join("\n"); }
